/*=============================================================================
	CAVectorUnitTypes.h
	
	$Log: CAVectorUnitTypes.h,v $
	Revision 1.1  2005/07/09 00:02:48  asynth
	add CAVectorUnitTypes.h
	

	created 07/08/05, James McCartney
	Copyright (c) 2005 Apple Computer, Inc.  All Rights Reserved

	$NoKeywords: $
=============================================================================*/

#ifndef __CAVectorUnitTypes_h__
#define __CAVectorUnitTypes_h__

enum {
	kVecUninitialized = -1,
	kVecNone = 0,
	kVecAltivec = 1,
	kVecSSE2 = 100,
	kVecSSE3 = 101
};

#endif
