/*=============================================================================
	CAXException.cpp
	
	$Log: CAXException.cpp,v $
	Revision 1.3  2004/01/13 02:00:44  dwyatt
	moved from Source/Tests/AudioFileUtility/Utility/
	
	Revision 1.2  2003/08/23 04:54:00  dwyatt
	add XThrow, support for warnings
	
	Revision 1.1  2003/06/23 23:07:37  dwyatt
	initial checkin
	
	created Wed Jun 11 2003, Doug Wyatt
	Copyright (c) 2003 Apple Computer, Inc.  All Rights Reserved

	$NoKeywords: $
=============================================================================*/

#include "CAXException.h"

CAXException::WarningHandler CAXException::sWarningHandler = NULL;
