/*=============================================================================
	CAConditionalMacros.h

	$Log: CAConditionalMacros.h,v $
	Revision 1.5  2004/10/08 01:33:05  jcm10
	account for __COREAUDIO_USE_FLAT_INCLUDES__
	
	Revision 1.4  2003/11/20 22:56:53  dwyatt
	__COREAUDIO_USE_FLAT_INCLUDES__
	
	Revision 1.3  2003/01/27 20:06:32  jcm10
	fix the typo
	
	Revision 1.2  2002/05/14 00:18:11  luke
	integrate changes made in SoundMgr respository <ealdrich, luke>
	
	Revision 1.1  2002/04/23 21:51:29  jcm10
	clean up things for the AudioCodec build
	
	Revision 1.1  2002/04/23 00:35:37  jcm10
	moved here from CAServices
	
	Revision 0.0  Mon Apr 22 2002 16:29:53 US/Pacific  moorf
	Created
		
	$NoKeywords: $
=============================================================================*/
#if !defined(__CAConditionalMacros_h__)
#define __CAConditionalMacros_h__

//=============================================================================
//	This file exists to make figuring out how to include system headers
//	easier in a cross platform world. We throw in an include of the standard
//	ConditionalMacros too.
//=============================================================================

// ########## THIS FILE SHOULD GO AWAY SOON, replaced by __COREAUDIO_USE_FLAT_INCLUDES__
// but for now, use this as a way to define __COREAUDIO_USE_FLAT_INCLUDES__ programmatically

//	TargetConditionals.h defines the bare minimum we need
#include "TargetConditionals.h"

//	Determine whether or not to use framework style includes for system headers
#if !defined(CoreAudio_Use_Framework_Includes) && !defined(__COREAUDIO_USE_FLAT_INCLUDES__)
	#if	TARGET_RT_MAC_MACHO
		#define	CoreAudio_Use_Framework_Includes	1
	#else
		#define	CoreAudio_Use_Framework_Includes	0
	#endif
#endif

//	Include the regular ConditionalMacros.h too, since it has useful stuff that
//	TargetConditionals.h lacks for some reason.
#if	CoreAudio_Use_Framework_Includes
	#include <CoreServices/../Frameworks/CarbonCore.framework/Headers/ConditionalMacros.h>
#else
	#include "ConditionalMacros.h"
#endif

#endif
