/*
 *  CABundleLocker.h
 *
 *  Created by James McCartney on Tue Dec 21 2004.
 *  Copyright (c) 2004 Apple Computer. All rights reserved.
 *
 *	$Log: CABundleLocker.h,v $
 *	Revision 1.2  2005/02/16 02:13:41  jcm10
 *	make it work for Windows
 *	
 *	Revision 1.1  2005/01/12 17:39:09  luke
 *	moved here from CAServices/Utility
 *	
 *	
 *	$NoKeywords: $
 */

#ifndef _CABundleLocker_h_
#define _CABundleLocker_h_


#include "CAConditionalMacros.h"

/*
some bundle operations are not thread safe, notably CFCopyLocalizedStringFromTableInBundle
*/

class CABundleLocker
{
public:

#if TARGET_OS_MAC
	CABundleLocker();
	~CABundleLocker();
#else
	CABundleLocker() {}
	~CABundleLocker() {}
#endif
};

#endif
